"""Demo of GUITk widgets and layouts"""


import contextlib
import pathlib
import sys
import tkinter as tk

import guitk
from guitk import *
from guitk.containers import _Container

SAMPLE_TEXT = """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Cras tempus neque in vehicula hendrerit. 
Nam non posuere ante. Nunc libero libero, tempus eget enim vitae, egestas hendrerit tortor. Vivamus et egestas felis. 
Aliquam erat volutpat. Nulla facilisi. Aliquam hendrerit, nibh nec tempor lobortis, purus nisl vehicula ex, dapibus 
fermentum mauris nibh id nulla. Vivamus non pretium quam. Phasellus elementum commodo nisl. Nullam eu faucibus augue. 
Vivamus pulvinar metus vehicula urna porttitor euismod. """
SAMPLE_TEXT = SAMPLE_TEXT.replace("\n", "")


ALL_WIDGETS = guitk.__all__


def get_docstring(name):
    """Return the docstring of an object with name"""
    try:
        obj = globals()[name]
    except KeyError as e:
        raise ValueError(f"Invalid object name: {e}") from e
    return obj.__doc__ or ""


def list_files(path: str, tree: Treeview):
    """list files in a directory and add to Treeview tree"""
    files = pathlib.Path(path).iterdir()
    pyfiles = []
    for f in files:
        if f.suffix == ".py":
            pyfiles.append(str(f))
            tags = ["pyfile"]
        else:
            tags = []
        tree.widget.insert(
            "",
            "end",
            iid=str(f),
            tags=tags,
            text=str(f),
            values=(str(f), f.stat().st_size),
        )
    tree.widget.selection_set(pyfiles)


class Demo(Window):
    def config(self):
        self.title = "GUTk Demo"
        self.create_layout()
        self.create_menubar()

    def create_layout(self):
        with VLayout():
            with HStack(halign=tk.CENTER):
                Label("This is a demo of GUITk widgets and layouts")
            HSeparator()
            with Notebook(sticky="nsew", weightx=1):
                with VTab("Tab 1"):
                    with HStack():
                        LabelEntry("File:", key="file")
                        BrowseFileButton(target_key="file")
                        LabelEntry("Directory:", key="directory")
                        BrowseDirectoryButton(target_key="directory")
                        Button("Debug Window", key="debug_window")
                    with HStack():
                        LabelEntry("Enter some text:", key="entry", keyrelease=True)
                        Label("You entered:")
                        Label("", key="label_entered")
                    with HStack():
                        LinkLabel("This Label is a link", key="link", sticky="ns").font(
                            underline=True
                        ).style(foreground="blue")
                        CheckButton(
                            "Enable other checkbutton",
                            key="check_enable_other",
                            sticky="ns",
                        )
                        CheckButton(
                            "Other checkbutton",
                            key="check_other",
                            disabled=True,
                            sticky="ns",
                        )
                        Button("Start Timer", key="start_timer", sticky="ns")
                        Label("0", key="timer", width=4, sticky="ns")
                        ComboBox(
                            default="PNG",
                            values=["PNG", "JPEG", "GIF"],
                            width=4,
                            readonly=True,
                        )
                        SpinBox(from_value=0, to_value=5, wrap=True, width=4)
                    with HStack():
                        with LabelFrame("Radio Buttons"):
                            with VStack():
                                Label("These RadioButtons are in a LabelFrame")
                                with HStack():
                                    with VStack(vexpand=False, hexpand=False):
                                        RadioButton(
                                            "Option A", key="option_a", group="group1"
                                        )
                                        RadioButton(
                                            "Option B", key="option_b", group="group1"
                                        )
                                    with VStack(vexpand=False, hexpand=False):
                                        RadioButton(
                                            "Option 1", key="option_1", group="group2"
                                        )
                                        RadioButton(
                                            "Option 2", key="option_2", group="group2"
                                        )
                        with VStack():
                            Label("These Text boxes are in an HStack")
                            with HStack():
                                with VStack():
                                    Label("Text Box", pady=0)
                                    Text(
                                        key="text_box_1",
                                        text=SAMPLE_TEXT,
                                        height=4,
                                        tooltip="Text box",
                                        pady=0,
                                    )
                                with VStack(padding=0):
                                    Label("Text Box With Scrollbar", pady=0)
                                    Text(
                                        key="text_box_2",
                                        text=SAMPLE_TEXT,
                                        height=4,
                                        vscrollbar=True,
                                        tooltip="Text box with scrollbar",
                                        pady=0,
                                    )
                    with HStack():
                        Treeview(
                            key="treeview",
                            headings=["Filename", "Size"],
                            show="headings",
                            tooltip="Treeview; click on headings to sort",
                            vscrollbar=True,
                        ),
                        with VStack(vexpand=False, hexpand=False, halign="center"):
                            Label("Classes and functions in GUITk")
                            ListBox(
                                key="listbox",
                                text=ALL_WIDGETS,
                                height=10,
                                width=160,
                                vscrollbar=True,
                            )
                        with VStack(vexpand=False, hexpand=False, halign="center"):
                            Label("Doc String")
                            Text(key="docstring", height=14, width=30, vscrollbar=True)
                with HTab("Tab 2"):
                    with PanedWindow(key="paned_window", weightx=1):
                        with HPane():
                            Text(text=SAMPLE_TEXT, weightx=1, width=None)
                        with VPane():
                            Text(text=SAMPLE_TEXT, weightx=1, width=None)
            with LabelFrame("Output Redirect", weightx=1, sticky="ew"):
                with HStack():
                    Output(key="output", sticky="nsew", weightx=1, weighty=1)
                    VSeparator()
                    with VStack(vexpand=False, hexpand=False, valign=tk.BOTTOM):
                        Checkbutton("Enable", key="check_enable")
                        Checkbutton("Echo", key="check_echo")
                        Button("stdout")
                        Button("stderr")

    def create_menubar(self):
        """create the menu"""
        with MenuBar():
            with Menu("File"):
                Command("Open", shortcut="Ctrl+O")
                Command("Save")
                Command("Save As")
                Command("Exit", command=self.quit)
            with Menu("Edit"):
                Command("Cut")
                Command("Copy")
                Command("Paste")
            with Menu("SubMenus"):
                with Menu("SubMenu 1"):
                    Command("Command 1")
                    Command("Command 2")
                with Menu("SubMenu 2"):
                    Command("Command 3")
                    Command("Command 4")

    def setup(self):
        """gets called right after __init__"""
        print(f"Setting up: {self.__dict__}")

        # check the Enable checkbutton for enabled output redirect
        self.get("check_enable").widget.invoke()

        # set up the treeview
        self.tree_sort_size_reverse = False
        self.tree_sort_filename_reverse = False
        tree = self.get("treeview")
        list_files(".", self.get("treeview"))

        # bind events for treeview
        tree.bind_heading("Size", "tree_heading_size")
        tree.bind_heading("Filename", "tree_heading_filename")

        # handle Return and double-click events for treeview, but only for python files
        # list_files() will tag python files with tag 'pyfile'
        tree.bind_tag("pyfile", "tree_python_file", sequence="<Return>")
        tree.bind_tag("pyfile", "tree_python_file", sequence="<Double-1>")

        # will store timer id for the timer
        self.timer_id = None

        # set up the listbox
        listbox = self.get("listbox")
        first_item = ALL_WIDGETS[0]
        listbox.widget.selection_set(first_item)
        listbox.widget.see(first_item)

    def teardown(self):
        """gets called right before window is destroyed"""
        print("Tearing down")

    def handle_event(self, event: Event):
        """Gets called for every event"""
        print(event)

    @on(key="debug_window")
    def on_debug_window(self):
        DebugWindow(parent=self.window)

    @on(key="entry")
    def on_entry(self):
        self["label_entered"].value = self["entry"].value

    @on(key="check_enable")
    def on_check_enable(self):
        if self["check_enable"].value:
            print("Enabling output redirect")
            self["output"].enable_redirect()
        else:
            print("Disabling output redirect")
            self["output"].disable_redirect()

    @on(key="check_echo")
    def on_check_echo(self):
        self["output"].echo = self["check_echo"].value

    @on(key="stdout")
    def on_stdout(self):
        print("This is stdout")

    @on(key="stderr")
    def on_stderr(self):
        print("This is stderr", file=sys.stderr)

    @on(key="link")
    def on_link(self):
        print("You clicked the link")

    @on(key="check_enable_other")
    def on_check_enable_other(self):
        self["check_other"].disabled = not self["check_enable_other"].value

    @on(key="start_timer")
    def on_start_timer(self):
        """Start/stop the timer"""
        # cancel the timer if it is running
        if self.timer_id:
            self.cancel_timer_event(self.timer_id)
            self.timer_id = None
            self.get("start_timer").value = "Start Timer"
            return

        # start the timer
        self.timer_id = self.bind_timer_event(1000, "<<MyTimer>>", repeat=True)
        self.get("start_timer").value = "Stop Timer"

    @on(key="<<MyTimer>>")
    def on_timer(self):
        """Increment the timer"""
        self["timer"].value = int(self["timer"].value) + 1

    @on(key="tree_heading_size")
    def on_tree_heading_size(self):
        """Sort treeview on size"""
        self.get("treeview").sort_on_column(
            "Size",
            reverse=self.tree_sort_size_reverse,
            key=lambda x: int(x[0]),
        )
        self.tree_sort_size_reverse = not self.tree_sort_size_reverse

    @on(key="tree_heading_filename")
    def on_tree_heading_filename(self):
        """Sort treeview on filename"""
        self.get("treeview").sort_on_column(
            "Filename",
            reverse=self.tree_sort_filename_reverse,
            key=lambda x: x[0],
        )
        self.tree_sort_filename_reverse = not self.tree_sort_filename_reverse

    @on(key="tree_python_file")
    def on_tree_python_file(self):
        """User hit return on a Python file"""
        print("You picked a Python file")

    @on(key="treeview")
    def on_treeview(self):
        """User clicked on a treeview item"""
        print(f"You clicked on a treeview item: {self.get('treeview').value}")

    @on(key="listbox")
    def on_listbox(self):
        """User clicked on a listbox item, update docstring text box"""
        docstring = None
        with contextlib.suppress(AttributeError):
            docstring = get_docstring(self.get("listbox").value[0])
        docstring = docstring or "doc string not found"
        self.get("docstring").value = docstring


if __name__ == "__main__":
    Demo().run()



""" Container classes for guitk """

from __future__ import annotations

import tkinter as tk
from typing import TYPE_CHECKING, Hashable

from guitk.constants import GUITK

from ._debug import debug, debug_borderwidth, debug_relief, debug_watch
from .basewidget import BaseWidget
from .frame import _Container
from .spacer import HSpacer, VSpacer
from .types import HAlign, PaddingType, PadType, VAlign

if TYPE_CHECKING:
    from .window import Window


class _Stack(_Container):
    """A container that stacks widgets when added to a Layout"""

    def __init__(
        self,
        key: Hashable | None = None,
        height: int | None = None,
        width: int | None = None,
        padding: PaddingType | None = None,
        disabled: bool | None = False,
        sticky: str | None = "nsew",
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        vexpand: bool = True,
        hexpand: bool = True,
        distribute: bool = False,
        vspacing: PadType | None = None,
        hspacing: PadType | None = None,
        vscrollbar: bool = False,
        # hscrollbar: bool = False,
        autohide_scrollbars: bool = True,
    ):
        """Base container container that stacks widgets when added to a Layout.

        The base container stacks widgets in a single vertical column. Subclasses
        can override the layout property to change the layout to horizontal or
        grid layouts.

        Args:
            key (Hashable, optional): The key to use for the Stack. Defaults to None.
            height (int, optional): The height of the Stack. Defaults to None.
            width (int, optional): The width of the Stack. Defaults to None.
            padding (PaddingType, optional): The padding around the Stack. Defaults to None.
            disabled (bool, optional): Whether the VStack is disabled. Defaults to False.
            sticky (str, optional): The sticky value for the Stack. Defaults to "nsew".
            valign (VAlign, optional): The vertical alignment for the widgets in the Stack.
                Defaults to None.
            halign (HAlign, optional): The horizontal alignment for the widgets in the Stack.
                Defaults to None.
            vexpand (bool, optional): Whether the Stack should expand vertically.
                Defaults to True.
            hexpand (bool, optional): Whether the Stack should expand horizontally.
                Defaults to True.
            distribute (bool, optional): Whether the Stack should distribute widgets evenly.
            vspacing (PadType, optional): Vertical spacing between widgets. Defaults to None.
            hspacing (PadType, optional): Horizontal spacing between widgets. Defaults to None.
            vscrollbar (bool): Whether to include a vertical scrollbar. Defaults to False.

        Note:
            If width or height is specified, the Stack will not expand to fill the available space and the
            expand parameter will be ignored.
        """
        super().__init__(
            frametype=GUITK.ELEMENT_FRAME,
            key=key,
            width=width,
            height=height,
            layout=None,
            style=None,
            borderwidth=debug_borderwidth() or None,
            padding=padding,
            relief=debug_relief() or None,
            disabled=disabled,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=False,
            padx=0,
            pady=0,
            valign=valign,
            halign=halign,
            vspacing=vspacing,
            hspacing=hspacing,
            vscrollbar=vscrollbar,
            # hscrollbar=hscrollbar,
            autohide_scrollbars=autohide_scrollbars,
        )
        self.vexpand = vexpand if height is None else False
        self.hexpand = hexpand if width is None else False
        self.distribute = distribute
        self._layout_list = []
        self._layout_lol = [[]]

    def _create_widget(self, parent: tk.BaseWidget, window: Window, row: int, col: int):
        super()._create_widget(parent, window, row, col)

        if self.vexpand:
            parent.grid_rowconfigure(row, weight=1)
        if self.hexpand:
            parent.grid_columnconfigure(col, weight=1)

    @property
    def layout(self) -> list[list[BaseWidget]]:
        """Return the layout of the Stack"""
        if self.distribute:
            self._layout_lol = []
            for widget in self._layout_list:
                self._layout_lol.append([VSpacer()])
                self._layout_lol.append([widget])
            self._layout_lol.append([VSpacer()])
        else:
            self._layout_lol = (
                [[widget] for widget in self._layout_list]
                if self._layout_list
                else [[]]
            )
        return self._layout_lol

    @layout.setter
    def layout(self, layout: list[list[BaseWidget]]):
        """Set the layout of the Stack"""
        self._layout_lol = layout
        self._layout_list = [widget for row in layout for widget in row]

    @property
    def widgets(self) -> list[BaseWidget]:
        """Return a list of the widgets in the Stack"""
        return self._layout_list or []

    def append(self, widget: BaseWidget):
        """Add a widget to the bottom of the Stack"""
        self._layout_list.append(widget)
        self.redraw()

    def extend(self, widgets: list[BaseWidget]):
        """Add a list of widgets to the end of the Stack"""
        for widget in widgets:
            self.append(widget)

    def insert(self, index: int, widget: BaseWidget):
        """Insert a widget at the given index in the Stack.

        Args:
            index (int): The index to insert the widget at.
            widget (Widget): The widget to insert.

        Note: The first argument is the index of the element before which to insert,
        so a.insert(0, x) inserts at the front of the stack, and a.insert(len(a), x)
        is equivalent to a.append(x).

        If the index is out of range, the widget will be added to the end of the Stack.
        """
        self._layout_list.insert(index, widget)
        self.redraw()

    def clear(self):
        """Remove all widgets from the Stack"""
        # copy the list so we can iterate over it while removing widgets
        # which will change the list
        for widget in self._layout_list.copy():
            debug(f"{self} destroying {widget=}")
            widget.destroy()
        self._layout_list = []
        self.redraw()

    def pop(self, index: int = -1):
        """Remove and return the widget at the given index in the Stack.

        Args:
            index (int): The index of the widget to remove.

        Returns:
            Widget: The widget that was removed.

        Raises:
            IndexError: If the index is out of range.
        """
        widget = self._layout_list.pop(index)
        widget.widget.grid_forget()
        self.redraw()
        return widget

    @debug_watch
    def remove(self, key_or_widget: Hashable | BaseWidget):
        """Remove widget from the container and destroy it.

        Args:
            key_or_widget (Hashable | Widget): The key of the widget to remove or the widget object. If a key is given,
                the first widget with that key will be removed.

        Raises:
            ValueError: If the widget is not in the Stack.
        """
        for idx, widget in enumerate(self._layout_list):
            if widget == key_or_widget or widget.key == key_or_widget:
                debug(
                    f"removing {key_or_widget} from {self} {widget.key} {widget.widget}"
                )
                widget = self._layout_list.pop(idx)
                self.window._forget_widget(widget)
                widget.widget.grid_forget()
                widget.widget.destroy()
                self.redraw()
                return
        raise ValueError(f"Widget {key_or_widget} not found in Stack")

    def redraw(self):
        """Redraw the Stack"""
        self._layout(self.frame, self.window)
        self.window.window.update_idletasks()

    def _add_widget(self, widget: BaseWidget):
        """Add a widget to the frame's layout"""
        self._layout_list.append(widget)

    @debug_watch
    def _insert_widget_row_col(
        self, widget: BaseWidget, row: int, col: int, is_vertical: bool = True
    ):
        """Insert a widget into the container after the container has been created
            Intended for use at run-time only when widgets need to be added dynamically

        Args:
            widget: (Widget) the widget to add
            row: (int) the row to insert the widget into
            col: (int) the column to insert the widget into

        Note: widgets are placed in a grid with row, col coordinates by the _LayoutMixin class.
        All containers store widgets internally as a list so we need to convert the row, col
        coordinates to a list index.
        """

        if is_vertical:
            # convert row, col to list index
            index = row
        else:
            index = col

        self._layout_list.insert(index, widget)

        # redraw the layout which will create the widget
        self._layout(self.frame, self.window)

    def __len__(self):
        """Length of the Stack (number of widgets contained)"""
        return len(self._layout_list) if self._layout_list else 0

    def __getitem__(self, index: int):
        """Get the widget at the given index"""
        return self._layout_list[index]

    def __delitem__(self, index: int):
        """Remove the widget at the given index"""
        self.pop(index)


class VStack(_Stack):
    """A container that stacks widgets vertically when added to a Layout"""

    def __init__(
        self,
        key: Hashable | None = None,
        width: int | None = None,
        # height: int | None = None,
        padding: PaddingType | None = None,
        disabled: bool | None = False,
        sticky: str | None = "nsew",
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        vexpand: bool = True,
        hexpand: bool = True,
        distribute: bool = False,
        vspacing: PadType | None = None,
        hspacing: PadType | None = None,
        vscrollbar: bool = False,
        # hscrollbar: bool = False,
        autohide_scrollbars: bool = True,
    ):
        """Base container container that stacks widgets vertically when added to a Layout

        Args:
            key (Hashable, optional): The key to use for the VStack. Defaults to None.
            width (int, optional): The width of the VStack. Defaults to None.
            padding (PaddingType, optional): The padding around the VStack. Defaults to None.
            disabled (bool, optional): Whether the VStack is disabled. Defaults to False.
            sticky (str, optional): The sticky value for the VStack. Defaults to "nsew".
            valign (VAlign, optional): The vertical alignment for the widgets in the VStack.
                Defaults to None.
            halign (HAlign, optional): The horizontal alignment for the widgets in the VStack.
                Defaults to None.
            vexpand (bool, optional): Whether the Stack should expand vertically.
                Defaults to True.
            hexpand (bool, optional): Whether the Stack should expand horizontally.
                Defaults to True.
            distribute (bool, optional): Whether the VStack should distribute widgets evenly.
            vspacing (PadType, optional): Vertical spacing between widgets. Defaults to None.
            hspacing (PadType, optional): Horizontal spacing between widgets. Defaults to None.
            vscrollbar (bool): Whether to include a vertical scrollbar. Defaults to False.
            autohide_scrollbars (bool): Whether to hide scrollbars when not needed. Defaults to True.

        Note:
            If width is specified, the VStack will not expand to fill the available space and the
            expand parameter will be ignored.
        """
        super().__init__(
            key=key,
            width=width,
            height=None,
            padding=padding,
            disabled=disabled,
            sticky=sticky,
            valign=valign,
            halign=halign,
            vexpand=vexpand,
            hexpand=hexpand,
            distribute=distribute,
            vspacing=vspacing,
            hspacing=hspacing,
            vscrollbar=vscrollbar,
            # hscrollbar=hscrollbar,
            autohide_scrollbars=autohide_scrollbars,
        )


class HStack(_Stack):
    """A container that stacks widgets horizontally when added to a Layout"""

    def __init__(
        self,
        key: Hashable | None = None,
        height: int | None = None,
        padding: PaddingType | None = None,
        disabled: bool | None = False,
        sticky: str | None = "nsew",
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        vexpand: bool = True,
        hexpand: bool = True,
        distribute: bool = False,
        vspacing: PadType | None = None,
        hspacing: PadType | None = None,
        vscrollbar: bool = False,
        # hscrollbar: bool = False,
        autohide_scrollbars: bool = True,
    ):
        """A container that stacks widgets horizontally when added to a Layout

        Args:
            key (Hashable, optional): The key to use for the HStack. Defaults to None.
            height (int, optional): The height of the HStack. Defaults to None.
            padding (PaddingType, optional): The padding around the HStack. Defaults to None.
            disabled (bool, optional): Whether the HStack is disabled. Defaults to False.
            sticky (str, optional): The sticky value for the HStack. Defaults to "nsew".
            valign (VAlign, optional): The vertical alignment for the widgets in the HStack.
                Defaults to None.
            halign (HAlign, optional): The horizontal alignment for the widgets in the HStack.
                Defaults to None.
            vexpand (bool, optional): Whether the Stack should expand vertically.
                Defaults to True.
            hexpand (bool, optional): Whether the Stack should expand horizontally.
                Defaults to True.
            distribute (bool, optional): Whether the HStack should distribute widgets evenly.
            vspacing (PadType, optional): Vertical spacing between widgets. Defaults to None.
            hspacing (PadType, optional): Horizontal spacing between widgets. Defaults to None.
            vscrollbar (bool): Whether to include a vertical scrollbar. Defaults to False.
            autohide_scrollbars (bool): Whether to hide scrollbars when not needed. Defaults to True.

        Note:
            If height is specified, the HStack will not expand to fill the available space and the
            expand parameter will be ignored.
        """
        super().__init__(
            key=key,
            height=height,
            width=None,
            padding=padding,
            disabled=disabled,
            sticky=sticky,
            valign=valign,
            halign=halign,
            vexpand=vexpand,
            hexpand=hexpand,
            distribute=distribute,
            vspacing=vspacing,
            hspacing=hspacing,
            vscrollbar=vscrollbar,
            # hscrollbar=hscrollbar,
            autohide_scrollbars=autohide_scrollbars,
        )

    @property
    def layout(self) -> list[list[BaseWidget]]:
        """Return the layout of the HStack"""
        if self.distribute:
            self._layout_lol = [[]]
            for widget in self._layout_list:
                self._layout_lol[0].append(HSpacer())
                self._layout_lol[0].append(widget)
            self._layout_lol[0].append(HSpacer())
        else:
            self._layout_lol = [self._layout_list]
        return self._layout_lol

    @layout.setter
    def layout(self, layout: list[list[BaseWidget]]):
        """Set the layout of the VStack"""
        self._layout_lol = layout
        self._layout_list = [widget for row in layout for widget in row]

    @debug_watch
    def _insert_widget_row_col(
        self, widget: BaseWidget, row: int, col: int, is_vertical: bool = True
    ):
        """Insert a widget into the container after the container has been created
            Intended for use at run-time only when widgets need to be added dynamically

        Args:
            widget: (Widget) the widget to add
            row: (int) the row to insert the widget into
            col: (int) the column to insert the widget into

        Note: widgets are placed in a grid with row, col coordinates by the _LayoutMixin class.
        All containers store widgets internally as a list so we need to convert the row, col
        coordinates to a list index.
        """
        super()._insert_widget_row_col(widget, row, col, is_vertical=False)


class VGrid(_Stack):
    """A container that arranges widgets in a vertical grid when added to a Layout"""

    def __init__(
        self,
        rows: int,
        key: Hashable | None = None,
        width: int | None = None,
        padding: PaddingType | None = None,
        disabled: bool | None = False,
        sticky: str | None = "nsew",
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        vexpand: bool = True,
        hexpand: bool = True,
        vspacing: PadType | None = None,
        hspacing: PadType | None = None,
        vscrollbar: bool = False,
        # hscrollbar: bool = False,
        autohide_scrollbars: bool = True,
    ):
        """Container that stacks widgets in a vertical grid when added to a Layout

        Args:
            rows (int): The number of rows in the grid. The number of columns is determined by the
                number of widgets added to the grid.
            key (Hashable, optional): The key to use for the VStack. Defaults to None.
            width (int, optional): The width of the VStack. Defaults to None.
            padding (PaddingType, optional): The padding around the VStack. Defaults to None.
            disabled (bool, optional): Whether the VStack is disabled. Defaults to False.
            sticky (str, optional): The sticky value for the VStack. Defaults to "nsew".
            valign (VAlign, optional): The vertical alignment for the widgets in the VStack.
                Defaults to None.
            halign (HAlign, optional): The horizontal alignment for the widgets in the VStack.
                Defaults to None.
            vexpand (bool, optional): Whether the Stack should expand vertically.
                Defaults to True.
            hexpand (bool, optional): Whether the Stack should expand horizontally.
                Defaults to True.
            vspacing (PadType, optional): Vertical spacing between widgets. Defaults to None.
            hspacing (PadType, optional): Horizontal spacing between widgets. Defaults to None.
            vscrollbar (bool): Whether to include a vertical scrollbar. Defaults to False.
            autohide_scrollbars (bool): Whether to hide scrollbars when not needed. Defaults to True.

        Note:
            If width is specified, the VStack will not expand to fill the available space and the
            expand parameter will be ignored.
        """
        super().__init__(
            key=key,
            height=None,
            width=width,
            padding=padding,
            disabled=disabled,
            sticky=sticky,
            valign=valign,
            halign=halign,
            vexpand=vexpand,
            hexpand=hexpand,
            distribute=False,
            vspacing=vspacing,
            hspacing=hspacing,
            vscrollbar=vscrollbar,
            # hscrollbar=hscrollbar,
            autohide_scrollbars=autohide_scrollbars,
        )
        self.rows = rows

    @property
    def layout(self) -> list[list[BaseWidget]]:
        """Return the layout of the Stack"""
        if self.distribute:
            raise NotImplementedError("distribute note yet implemented for VGrid")
        else:
            # grid the widgets in the order they were added
            self._layout_lol = [
                self._layout_list[i :: self.rows] for i in range(self.rows)
            ]
        return self._layout_lol

    @layout.setter
    def layout(self, layout: list[list[BaseWidget]]):
        """Set the layout of the Stack"""
        ...

    @debug_watch
    def _insert_widget_row_col(
        self, widget: BaseWidget, row: int, col: int, is_vertical: bool = True
    ):
        """Insert a widget into the container after the container has been created
            Intended for use at run-time only when widgets need to be added dynamically

        Args:
            widget: (Widget) the widget to add
            row: (int) the row to insert the widget into
            col: (int) the column to insert the widget into

        Note: widgets are placed in a grid with row, col coordinates by the _LayoutMixin class.
        All containers store widgets internally as a list so we need to convert the row, col
        coordinates to a list index.
        """
        index = self._row_col_to_index(row, col)
        self._layout_list.insert(index, widget)

        # redraw the layout which will create the widget
        self._layout(self.frame, self.window)

    def _row_col_to_index(self, row: int, col: int):
        """Convert row, col coordinates to a list index"""
        # the VGrid layout converts a list into a list of lists of length self.rows
        # convert row, col to back to a list index
        return row + col * self.rows


class HGrid(_Stack):
    """A container that arranges widgets in a horizontal grid when added to a Layout"""

    def __init__(
        self,
        cols: int,
        key: Hashable | None = None,
        width: int | None = None,
        padding: PaddingType | None = None,
        disabled: bool | None = False,
        sticky: str | None = "nsew",
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        vexpand: bool = True,
        hexpand: bool = True,
        vspacing: PadType | None = None,
        hspacing: PadType | None = None,
        vscrollbar: bool = False,
        # hscrollbar: bool = False,
        autohide_scrollbars: bool = True,
    ):
        """Container that stacks widgets in a horizontal grid when added to a Layout

        Args:
            cols (int): The number of columns in the grid. The number of rows is determined by the
                number of widgets added to the grid.
            key (Hashable, optional): The key to use for the VStack. Defaults to None.
            width (int, optional): The width of the VStack. Defaults to None.
            padding (PaddingType, optional): The padding around the VStack. Defaults to None.
            disabled (bool, optional): Whether the VStack is disabled. Defaults to False.
            sticky (str, optional): The sticky value for the VStack. Defaults to "nsew".
            valign (VAlign, optional): The vertical alignment for the widgets in the VStack.
                Defaults to None.
            halign (HAlign, optional): The horizontal alignment for the widgets in the VStack.
                Defaults to None.
            vexpand (bool, optional): Whether the Stack should expand vertically.
                Defaults to True.
            hexpand (bool, optional): Whether the Stack should expand horizontally.
                Defaults to True.
            vspacing (PadType, optional): Vertical spacing between widgets. Defaults to None.
            hspacing (PadType, optional): Horizontal spacing between widgets. Defaults to None.
            vscrollbar (bool): Whether to include a vertical scrollbar. Defaults to False.
            autohide_scrollbars (bool): Whether to hide scrollbars when not needed. Defaults to True.

        Note:
            If width is specified, the VStack will not expand to fill the available space and the
            expand parameter will be ignored.
        """
        super().__init__(
            key=key,
            height=None,
            width=width,
            padding=padding,
            disabled=disabled,
            sticky=sticky,
            valign=valign,
            halign=halign,
            vexpand=vexpand,
            hexpand=hexpand,
            distribute=False,
            vspacing=vspacing,
            hspacing=hspacing,
            vscrollbar=vscrollbar,
            # hscrollbar=hscrollbar,
            autohide_scrollbars=autohide_scrollbars,
        )
        self.cols = cols

    @property
    def layout(self) -> list[list[BaseWidget]]:
        """Return the layout of the Stack"""
        if self.distribute:
            raise NotImplementedError("distribute not yet implemented for HGrid")
        else:
            # grid the widgets in the order they were added
            # chunk items into lists self.cols long
            self._layout_lol = [
                self._layout_list[i : i + self.cols]
                for i in range(0, len(self._layout_list), self.cols)
            ]
        return self._layout_lol

    @layout.setter
    def layout(self, layout: list[list[BaseWidget]]):
        """Set the layout of the Stack"""
        ...

    @debug_watch
    def _insert_widget_row_col(
        self, widget: BaseWidget, row: int, col: int, is_vertical: bool = True
    ):
        """Insert a widget into the container after the container has been created
            Intended for use at run-time only when widgets need to be added dynamically

        Args:
            widget: (Widget) the widget to add
            row: (int) the row to insert the widget into
            col: (int) the column to insert the widget into

        Note: widgets are placed in a grid with row, col coordinates by the _LayoutMixin class.
        All containers store widgets internally as a list so we need to convert the row, col
        coordinates to a list index.
        """
        index = self._row_col_to_index(row, col)
        self._layout_list.insert(index, widget)

        # redraw the layout which will create the widget
        self._layout(self.frame, self.window)

    def _row_col_to_index(self, row: int, col: int):
        """Convert row, col coordinates to a list index"""
        # the HGrid layout converts a list into a list of lists where each list is length self.cols
        # convert row, col to back to a list index
        return row * self.cols + col



"""Custom types for guitk"""

from __future__ import annotations

import tkinter as tk
from typing import Any, Callable, Literal, TypeVar

Widget = TypeVar("Widget")
TooltipType = Callable[[str], str | None]
CommandType = Callable[[], Any]
ValueType = tk.Variable
LayoutType = list[list[Widget | None]]
Window = TypeVar("Window")
VAlign = Literal["top", "bottom", "center"]
HAlign = Literal["left", "right", "center"]
DecoratedType = TypeVar("DecoratedType")
SizeType = tuple[int, int] | str | None
PaddingType = tuple[int, int, int, int] | tuple[int, int] | int | str
PadType = tuple[int, int] | int
CompoundType = (
    Literal["image", "text", "top", "bottom", "left", "right", "center"] | None
)


class _WindowBaseClass:
    """Just to keep typing happy"""

    ...



"""ttk Entry widget"""

from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType
from .utils import scrolled_widget_factory

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Entry", "LabelEntry"]

_valid_standard_attributes = {
    "class",
    "cursor",
    "style",
    "takefocus",
    "xscrollcommand",
}

_valid_ttk_entry_attributes = {
    "exportselection",
    "invalidcommand",
    "justify",
    "show",
    "state",
    "textvariable",
    "validate",
    "validatecommand",
    "width",
} | _valid_standard_attributes


class Entry(BaseWidget):
    """ttk.Entry text entry / input box"""

    def __init__(
        self,
        key: Hashable | None = None,
        default: str | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        keyrelease: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        hscrollbar: bool = False,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """Initialize an Entry widget.

        Args:
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            default (str | None, optional): Default text for the entry box. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to True.
            keyrelease (bool, optional): If True, generate events on key release. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            hscrollbar (bool, optional): Show horizontal scrollbar. Defaults to False.
            weightx (int | None, optional): Weight for horizontal resizing. Defaults to None.
            weighty (int | None, optional): Weight for vertical resizing. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Entry.

        Note:
            Emits EventType.EntryReturn event on return key press.
            If keyrelease is True, emits EventType.KeyRelease event on every key release.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            columnspan=columnspan,
            rowspan=rowspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.widget_type = "ttk.Entry"
        default = default or ""
        self._value.set(default)
        self.key = key or "Entry"
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.hscrollbar = hscrollbar
        self.keyrelease = keyrelease
        self.kwargs = kwargs

    def _create_widget(self, parent, window: Window, row, col):
        # build arg list for ttk.Entry
        kwargs_entry = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_entry_attributes
        }
        self.widget = scrolled_widget_factory(
            parent,
            ttk.Entry,
            hscrollbar=self.hscrollbar,
            textvariable=self._value,
            **kwargs_entry,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        # bind key release event
        if self.keyrelease:
            event = Event(self, window, self.key, EventType.KeyRelease)
            self.widget.bind("<KeyRelease>", window._make_callback(event))

        # bind return key event
        entry_return_key = Event(self, window, self.key, EventType.EntryReturn)
        self.widget.bind("<Return>", window._make_callback(entry_return_key))

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.KeyRelease,
                    command=self._command,
                )
            )

        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def entry(self):
        """Return the Tk entry widget"""
        return self.widget


class _ttkLabelEntry(ttk.Entry):
    """ttk.Entry with a Label"""

    def __init__(self, master=None, text=None, **kwargs):
        """Initialize a _ttkLabelEntry widget.

        Args:
            master (tk.Widget, optional): Parent widget. Defaults to None.
            text (str, optional): Label text. Defaults to None.
            **kwargs: Additional keyword arguments are passed to ttk.Entry.
        """
        self.frame = ttk.Frame(master)
        ttk.Entry.__init__(self, self.frame, **kwargs)
        self.label = ttk.Label(self.frame, text=text)
        self.label.grid(row=0, column=0)
        self.grid(row=0, column=1)

        # Copy geometry methods of self.frame without overriding Entry
        # methods -- hack!
        text_meths = vars(ttk.Entry).keys()
        methods = vars(tk.Pack).keys() | vars(tk.Grid).keys() | vars(tk.Place).keys()
        methods = methods.difference(text_meths)

        for m in methods:
            if m[0] != "_" and m not in {"config", "configure", "focus"}:
                setattr(self, m, getattr(self.frame, m))

    def __str__(self):
        return str(self.frame)


class LabelEntry(Entry):
    """Text entry / input box with a label"""

    # TODO: add option to put label above the entry box

    def __init__(
        self,
        text: str,
        key: Hashable | None = None,
        default: str | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        keyrelease: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        hscrollbar: bool = False,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """Initialize an Entry widget.

        Args:
            text (str): Label text.
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            default (str | None, optional): Default text for the entry box. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to True.
            keyrelease (bool, optional): If True, emits EventType.KeyRelease event on every key release.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            hscrollbar (bool, optional): Show horizontal scrollbar. Defaults to False.
            weightx (int | None, optional): Weight for horizontal resizing. Defaults to None.
            weighty (int | None, optional): Weight for vertical resizing. Defaults to None.
            focus (bool, optional): If True, widget will have focus. Defaults to False. Only one widget can have focus.
            **kwargs: Additional keyword arguments are passed to ttk.Entry.

        Note:
            Emits EventType.EntryReturn event on return key press.
            If keyrelease is True, emits EventType.KeyRelease event on every key release.
        """
        super().__init__(
            key=key,
            default=default,
            disabled=disabled,
            columnspan=columnspan,
            rowspan=rowspan,
            padx=padx,
            pady=pady,
            events=events,
            keyrelease=keyrelease,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            hscrollbar=hscrollbar,
            focus=focus,
            weightx=weightx,
            weighty=weighty,
        )
        self.widget_type = "guitk.LabelEntry"
        self.text = text
        self.kwargs = kwargs

    def _create_widget(self, parent, window: "Window", row, col):
        # build arg list for Entry
        kwargs_entry = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_entry_attributes
        }
        self.widget = _ttkLabelEntry(
            parent, text=self.text, textvariable=self._value, **kwargs_entry
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        # bind key release event
        event = Event(self, window, self.key, EventType.KeyRelease)
        self.widget.bind("<KeyRelease>", window._make_callback(event))

        # bind return key event
        entry_return_key = Event(self, window, self.key, EventType.EntryReturn)
        self.widget.bind("<Return>", window._make_callback(entry_return_key))

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.KeyRelease,
                    command=self._command,
                )
            )

        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def entry(self):
        """Return the Tk entry widget"""
        return self.widget



"""Redirect stdout/stderr """

from __future__ import annotations

import sys


class _StdOutRedirectBaseClass:
    """Base class for StdOutRedirect and StdErrRedirect"""

    def __init__(self):
        self._echo = False
        self._listeners = {}
        self._listener_count = 0
        self._redirect = False

        self.stdout = False
        self.stderr = False

    @property
    def echo(self):
        return self._echo

    @echo.setter
    def echo(self, val):
        self._echo = val

    def register(self, listener):
        if not callable(listener):
            raise ValueError("listener must be callable")
        self._listener_count += 1
        listener_id = self._listener_count
        self._listeners[listener_id] = listener
        self.enable_redirect()
        return listener_id

    def deregister(self, listener_id):
        del self._listeners[listener_id]
        if not self._listeners:
            self.disable_redirect()

    def write(self, line):
        if self._redirect:
            for listener in self._listeners:
                self._listeners[listener](line)
            if self.echo:
                if self.stdout:
                    sys.__stdout__.write(line)
                if self.stderr:
                    sys.__stderr__.write(line)

    def flush(self):
        pass

    def disable_redirect(self):
        self._redirect = False
        if self.stdout:
            sys.stdout = sys.__stdout__
        if self.stderr:
            sys.stderr = sys.__stderr__

    def enable_redirect(self):
        self._redirect = True
        if self.stdout:
            sys.stdout = self
        if self.stderr:
            sys.stderr = self

    def __del__(self):
        if self.stdout:
            sys.stdout = sys.__stdout__
        if self.stderr:
            sys.stderr = sys.__stderr__


class StdOutRedirect(_StdOutRedirectBaseClass):
    """Singleton class that handles redirect"""

    def __new__(cls, *args, **kwargs):
        """create new object or return instance of already created singleton"""
        if not hasattr(cls, "instance") or not cls.instance:
            cls.instance = super().__new__(cls)

        return cls.instance

    def __init__(self):
        if hasattr(self, "_init"):
            return
        super().__init__()
        self._init = True
        self.stdout = True


class StdErrRedirect(_StdOutRedirectBaseClass):
    """Singleton class that handles redirect"""

    def __new__(cls, *args, **kwargs):
        """create new object or return instance of already created singleton"""
        if not hasattr(cls, "instance") or not cls.instance:
            cls.instance = super().__new__(cls)

        return cls.instance

    def __init__(self):
        if hasattr(self, "_init"):
            return
        super().__init__()
        self._init = True
        self.stderr = True



"""Debug utilities; set GUITK_DEBUG=1 to enable debugging or use set_debug()"""

from __future__ import annotations

import datetime
import inspect
import os
import sys
import time
from functools import wraps

_global_debug = False

__all__ = [
    "debug",
    "debug_borderwidth",
    "debug_relief",
    "debug_watch",
    "is_debug",
    "set_debug",
]


def set_debug(value: bool):
    """Enable or disable debug"""
    global _global_debug
    _global_debug = value


def is_debug():
    """Return True if debug is enabled or False otherwise"""
    global _global_debug
    return _global_debug


def debug(*args, **kwargs):
    """Print debug message only if is_debug() is True"""
    if not is_debug():
        return

    # get caller name
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])
    caller = module.__name__

    # get line number of caller
    caller += f":{frame.lineno}"

    print(f"[DEBUG] {caller}: ", end="")
    print(*args, **kwargs)


def debug_borderwidth():
    """Return border width is_debug() is True or None otherwise"""
    return 1 if is_debug() else None


def debug_relief():
    """Return relief for frames if is_debug() is True or None otherwise"""
    return "raised" if is_debug() else None


def debug_watch(func):
    """decorate a function to print debug info before and after the function is called if debugging is enabled"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not is_debug():
            return func(*args, **kwargs)

        caller = sys._getframe().f_back.f_code.co_name
        name = func.__name__
        timestamp = datetime.datetime.now().isoformat()
        print(
            f"{timestamp} {name} called from {caller} with args: {args} and kwargs: {kwargs}"
        )
        start_t = time.perf_counter()
        rv = func(*args, **kwargs)
        stop_t = time.perf_counter()
        print(
            f"{timestamp} {name} returned: {rv}, elapsed time: {stop_t - start_t} sec"
        )
        return rv

    return wrapper


if os.getenv("GUITK_DEBUG") == "1":
    set_debug(True)
    debug("debugging enabled")



"""_LayoutMixin, _Container, and Frame classes"""

from __future__ import annotations

import contextlib
import tkinter as tk
from tkinter import ttk
from typing import TYPE_CHECKING, Hashable

from guitk.constants import GUITK

from ._debug import debug, debug_borderwidth, debug_relief, debug_watch
from .basewidget import BaseWidget
from .layout import HLayout, VLayout, pop_parent, push_parent
from .scrolledframe import ScrolledFrame
from .spacer import HSpacer, VSpacer
from .tooltips import Hovertip
from .ttk_label import Label
from .ttk_separator import HSeparator, VSeparator
from .types import HAlign, LayoutType, PaddingType, PadType, TooltipType, VAlign

_valid_frame_attributes = {
    "cursor",
    "height",
    "padding",
    "relief",
    "style",
    "takefocus",
    "width",
}

if TYPE_CHECKING:
    from .window import Window


def rows_columns(layout: LayoutType) -> tuple[int, int]:
    """Return the number of rows and columns in a layout"""
    rows = 0
    cols = 0
    for row in layout:
        rows += 1
        cols = max(cols, len(row))
    return rows, cols


class _LayoutMixin:
    """Mixin class to provide layout; for internal use only"""

    row_count: int = 0
    col_count: int = 0
    layout = []
    distribute: bool = False
    vspacing: PadType | None = None
    hspacing: PadType | None = None

    @debug_watch
    def _layout(self, parent: tk.BaseWidget, window: Window):
        """Create widgets from layout"""
        # as this is a mixin, make sure class being mixed into has necessary attributes

        # get alignment from layout or from class attributes
        debug(self)

        valign = "top"
        halign = "left"

        # first get alignment from layout
        with contextlib.suppress(AttributeError):
            valign = self.layout.valign.lower() if self.layout.valign else valign
            halign = self.layout.halign.lower() if self.layout.halign else halign

        # if widget has alignment attributes (e.g. VStack, HStack...), use those instead
        with contextlib.suppress(AttributeError):
            valign = self.valign.lower() if self.valign else valign
            halign = self.halign.lower() if self.halign else halign

        debug(f"valign={valign}, halign={halign}")

        if isinstance(self.layout, (HLayout, VLayout)):
            # layout may have been created outside of a Window
            # so ensure the layout has a reference to the Window
            self.layout.window = window

        layout = list(self.layout)
        rows, columns = rows_columns(layout)
        debug(f"{layout=} {rows=}, {columns=}")

        layout = self._add_spacers(layout, valign, halign)
        rows, columns = rows_columns(layout)
        debug(f"_add_spacers(): {layout=} {rows=}, {columns=}")

        for row_count, row in enumerate(layout):
            # if self.autoframe and (
            #     len(row) != 1
            #     or row[0].widget_type not in {"ttk.Frame", "tk.Frame", "LabelFrame"}
            # ):
            #     row_ = [
            #         _Container(layout=[row], autoframe=False, sticky="nsew", weightx=1)
            #     ]
            # else:
            #     row_ = row
            row_ = row
            self.row_count = row_count
            for col_count, widget in enumerate(row_):
                if widget is None:
                    # add blank label to maintain column spacing
                    # widget = Label("", disabled=True, events=False)
                    continue
                self._stickyfy(widget, valign, halign)
                debug(
                    f"{widget=}, {row_count=}, {col_count=}, {widget.sticky=} {widget.weightx=} {widget.weighty=}"
                )
                self._create_and_add_widget(
                    widget, parent, window, row_count, col_count
                )
                self.col_count = col_count

        debug(f"{self.row_count=}, {self.col_count=}")

    def _add_spacers(
        self, layout: LayoutType, valign: VAlign, halign: HAlign
    ) -> LayoutType:
        """Add spacers to layout"""

        rows, columns = rows_columns(layout)
        orientation = "vertical" if rows > 1 else "horizontal"
        new_layout = []
        for row in layout:
            new_row = row.copy()
            widget_idx = -1
            if halign in {"right", "center"}:
                if widget := new_row[0]:
                    # not None
                    if not widget.weightx:
                        new_row.insert(0, HSpacer())
                    else:
                        widget.columnspan = (
                            widget.columnspan + 1 if widget.columnspan else 2
                        )
                        new_row.append(None)
                        widget_idx = 0
            if halign == "center":
                if widget := new_row[widget_idx]:
                    if not widget.weightx:
                        new_row.append(HSpacer())
                    else:
                        widget.columnspan = widget.columnspan + 1
                        new_row.append(None)

            new_layout.append(new_row)

        if orientation == "horizontal":
            if valign in {"bottom", "center"}:
                # there's only one row so add spacers above/below as needed
                spacer_row = []
                for idx, widget in enumerate(new_layout[0]):
                    if not widget.weighty:
                        spacer_row.append(VSpacer())
                    else:
                        spacer_row.append(widget)
                        rowspan = 1 if valign == "bottom" else 2
                        widget.rowspan = (
                            widget.rowspan + rowspan if widget.rowspan else rowspan + 1
                        )
                        new_layout[0][idx] = None
                new_layout.insert(0, spacer_row)
            if valign in {"center"}:
                spacer_row = []
                for widget in new_layout[-1]:
                    if widget is None or widget.weighty is not None:
                        spacer_row.append(None)
                    else:
                        spacer_row.append(VSpacer())
                new_layout.append(spacer_row)
        else:
            rows, columns = rows_columns(new_layout)
            if valign in {"bottom", "center"}:
                new_layout.insert(
                    0,
                    [VSpacer(columnspan=columns), *[None for _ in range(columns - 1)]],
                )
            if valign in {"center"}:
                new_layout.append(
                    [
                        VSpacer(columnspan=columns),
                        *[None for _ in range(columns - 1)],
                    ]
                )

        return new_layout

    def _stickyfy(self, widget: BaseWidget, valign: str, halign: str):
        """Add correct stickiness to widget to achieve alignment"""
        sticky = ""
        weightx = 0
        weighty = 0
        if valign in {"top"}:
            sticky += "n"
        if valign in {"bottom"}:
            sticky += "s"
        if halign in {"right"}:
            sticky += "e"
        if halign in {"left"}:
            sticky += "w"
        if not widget.sticky:
            widget.sticky = sticky

    def _create_and_add_widget(self, widget, parent, window, row, col):
        """Create the widget and add it to layout"""

        if not widget._has_been_created:
            widget._set_parent_window(parent, window)
            # create the widget
            widget.key = widget.key or f"{widget.widget_type},{row},{col}"
            widget._create_widget(parent, window, row, col)

            # add tooltip if needed
            if tooltip := widget.tooltip or window.tooltip:
                _tooltip = tooltip(widget.key) if callable(tooltip) else tooltip
                widget._tooltip = (
                    Hovertip(widget.widget, _tooltip) if _tooltip else None
                )
            else:
                widget._tooltip = None

            window._widgets.append(widget)
            widget.parent = self
            window._widget_by_key[widget.key] = widget

            widget._has_been_created = True
        else:
            # grid the widget
            widget._grid(
                row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
            )

        widget._set_row_col(row, col)
        self._configure_widget(widget, parent, window, row, col)

    @debug_watch
    def _configure_widget(
        self,
        widget: BaseWidget,
        parent: tk.BaseWidget,
        window: Window,
        row: int,
        col: int,
    ):
        """Configure the widget"""

        # configure style if needed
        if widget._style_kwargs:
            style = ttk.Style()
            style_name = f"{id(widget)}.{widget.widget.winfo_class()}"
            style.configure(style_name, **widget._style_kwargs)
            widget.widget.configure(style=style_name)

        # configure row/columns/weight
        row_offset = 0
        col_offset = 0

        if widget.rowspan and widget.rowspan > 1:
            row_offset += widget.rowspan - 1
        if widget.columnspan and widget.columnspan > 1:
            col_offset += widget.columnspan - 1

        if widget.weightx is not None:
            parent.grid_columnconfigure(col + col_offset, weight=widget.weightx)
            # if widget created with scrolled_widget_factory
            # then need to configure the inner widget
            if getattr(widget.widget, "_guitk_framed_widget", False):
                widget.widget.grid_columnconfigure(0, weight=widget.weightx)
        if widget.weighty is not None:
            parent.grid_rowconfigure(row + row_offset, weight=widget.weighty)
            # configure inner widget if needed
            if getattr(widget.widget, "_guitk_framed_widget", False):
                widget.widget.grid_rowconfigure(0, weight=widget.weighty)

        # configure padding
        # if container has vspacing/hspacing and widget does not, use container's values
        if self.vspacing is not None and widget.pady is None:
            widget.pady = self.vspacing
        if self.hspacing is not None and widget.padx is None:
            widget.padx = self.hspacing
        padx = widget.padx if widget.padx is not None else window.padx
        pady = widget.pady if widget.pady is not None else window.pady
        widget.widget.grid_configure(padx=padx, pady=pady)

        # take focus if needed
        if widget._focus:
            widget.widget.focus()


class _Container(BaseWidget, _LayoutMixin):
    """Container base class for Frame and other containers; intended for internal use only"""

    def __init__(
        self,
        frametype: GUITK = GUITK.ELEMENT_FRAME,
        key: Hashable | None = None,
        layout: LayoutType | None = None,
        height: int | None = None,
        width: int | None = None,
        style: str | None = None,
        borderwidth: int | None = None,
        padding: PaddingType | None = None,
        relief: str = None,
        disabled: bool | None = False,
        rowspan: int | None = None,
        columnspan: int | None = None,
        text: str | None = None,
        labelanchor: str | None = None,
        sticky: str | None = None,
        tooltip: TooltipType | None = None,
        autoframe: bool | None = True,
        padx: PadType | None = None,
        pady: PadType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        vspacing: PadType | None = None,
        hspacing: PadType | None = None,
        vscrollbar: bool = False,
        # hscrollbar: bool = False,
        autohide_scrollbars: bool = True,
        **kwargs,
    ):
        # padx and pady passed to Widget not Frame
        BaseWidget.__init__(
            self,
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            sticky=sticky,
            tooltip=tooltip,
            padx=padx,
            pady=pady,
            weightx=weightx,
            weighty=weighty,
        )
        _LayoutMixin.__init__(self)

        # if autoframe is True, each widget or row of widgets will be placed in a frame
        self.autoframe = autoframe

        if frametype not in {
            GUITK.ELEMENT_FRAME,
            GUITK.ELEMENT_LABEL_FRAME,
            GUITK.ELEMENT_TK_FRAME,
        }:
            raise ValueError(f"bad frametype: {frametype}")
        self.frametype = frametype
        self.widget_type = frametype.value
        self.key = key or self.widget_type
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.width = width
        self.height = height
        self._style = style  # Widget has style() method so can't use self.style
        self.borderwidth = debug_borderwidth() or borderwidth
        self.padding = padding
        self.relief = debug_relief() or relief
        self.layout = layout or [[]]
        self.text = text
        self.labelanchor = labelanchor or "nw"
        self.kwargs = kwargs
        self.valign = valign.lower() if valign else None
        self.halign = halign.lower() if halign else None
        self.vspacing = vspacing
        self.hspacing = hspacing
        self.vscrollbar = vscrollbar
        # self.hscrollbar = hscrollbar
        self.autohide = autohide_scrollbars

    def remove(self, key_or_widget: Hashable | BaseWidget):
        """ "Remove widget from layout and destroy it.

        Args:
            key_or_widget (Hashable | Widget): The key or widget to remove. If a key is given,
                the first widget with that key will be removed.

        Raises:
            ValueError: If the widget is not found in the layout.
        """
        for row in self.layout:
            for idx, widget in enumerate(row):
                if widget == key_or_widget or widget.key == key_or_widget:
                    widget = row.pop(idx)
                    self.window._forget_widget(widget)
                    widget.widget.grid_forget()
                    widget.widget.destroy()
                    return
        raise ValueError(f"Widget not found: {key_or_widget}")

    def _create_widget(self, parent, window: Window, row, col):
        kwargs = {
            k: v
            for k, v in self.kwargs.items()
            if k in _valid_frame_attributes and v is not None
        }
        if self._style is not None:
            kwargs["style"] = self._style

        if self.frametype == GUITK.ELEMENT_FRAME:
            # if self.vscrollbar or self.hscrollbar:
            if self.vscrollbar:
                self.widget = ScrolledFrame(
                    parent,
                    vscrollbar=self.vscrollbar,
                    # hscrollbar=self.hscrollbar,
                    autohide=self.autohide,
                    width=self.width,
                    height=self.height,
                    borderwidth=self.borderwidth,
                    relief=self.relief,
                    **kwargs,
                )
            else:
                self.widget = ttk.Frame(
                    parent,
                    width=self.width,
                    height=self.height,
                    borderwidth=self.borderwidth,
                    relief=self.relief,
                    **kwargs,
                )
        elif self.frametype == GUITK.ELEMENT_TK_FRAME:
            self.widget = tk.Frame(
                parent,
                width=self.width,
                height=self.height,
                borderwidth=self.borderwidth,
                padx=self.padx,
                pady=self.pady,
                **kwargs,
            )
        else:
            self.widget = ttk.LabelFrame(
                parent,
                text=self.text,
                width=self.width,
                height=self.height,
                labelanchor=self.labelanchor,
                borderwidth=self.borderwidth,
                relief=self.relief,
                **kwargs,
            )

        if self.padding is not None and self.frametype != GUITK.ELEMENT_TK_FRAME:
            self.widget.configure(padding=self.padding)
        if self.relief is not None:
            self.widget["relief"] = self.relief
        if self.borderwidth is not None:
            self.widget["borderwidth"] = self.borderwidth

        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        if self.layout:
            self._layout(self.frame, window)

        if self.width or self.height:
            debug(f"{self.width=} {self.height=}")
            self.widget.grid_propagate(False)

        if self._disabled:
            self.widget.state(["disabled"])
        return self.widget

    def _insert_widget_row_col(
        self, widget: BaseWidget, row: int, col: int, vertical: bool = False
    ):
        """Insert a widget into the container after the container has been created
            Intended for use at run-time only when widgets need to be added dynamically

        Args:
            widget: (Widget) the widget to add
            row: (int) the row to insert the widget into
            col: (int) the column to insert the widget into
            vertical: (bool) if True, insert the widget vertically
        """
        # add widget to self.layout
        if vertical:
            row = min(row, len(self.layout))
            self._ensure_layout_size(row - 1, col or None)
            self.layout.insert(row, [widget])
        else:
            # python list.insert() will insert at the end if index is greater than the length
            # so copy that behavior here
            col = min(col, len(self.layout[row]))
            self._ensure_layout_size(row, None)
            self.layout[row].insert(col, widget)

        # redraw the layout which will create the widget
        self._layout(self.frame, self.window)

    def _ensure_layout_size(self, row: int, col: int | None):
        """Ensure the layout is at least row x col in size

        Args:
            row: (int) the row size to ensure
            col: (int or None) the column size to ensure; if None, does not add None to columns
        """
        while len(self.layout) <= row:
            self.layout.append([])
        if col is not None:
            while len(self.layout[row]) <= col:
                self.layout[row].append(None)

    def _pop_widget_row_col(
        self, row: int, col: int, vertical: bool = False
    ) -> BaseWidget:
        """Remove a widget from the container after the container has been created and return it
            Intended for use at run-time only when widgets need to be added dynamically

        Args:
            row: (int) the row to remove the widget from
            col: (int) the column to remove the widget from
            vertical: (bool) if True, remove the widget vertically

        Returns:
            Widget: the widget that was removed

        Raises:
            IndexError: if the row or column is out of range
        """
        try:
            widget = self.layout[row].pop(col)
            if vertical and not self.layout[row]:
                self.layout.pop(row)
        except IndexError as e:
            raise IndexError(f"No widget at row={row}, col={col}") from e

        widget.widget.grid_forget()
        self._layout(self.frame, self.window)

        return widget

    @property
    def frame(self):
        """Return the Tk Frame widget"""
        return self.widget

    @property
    def value(self):
        pass

    @value.setter
    def value(self, value):
        pass

    def _add_widget(self, widget: BaseWidget):
        """Add a widget to the frame's layout"""
        self.layout[0].append(widget)

    def __enter__(self):
        push_parent(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pop_parent()
        return False


class _VerticalContainer(_Container):
    """A Container class that lays out widgets vertically"""

    def __exit__(self, exc_type, exc_val, exc_tb):
        # reorder the layout to be vertical
        self.layout = [[x] for x in self.layout[0]]
        pop_parent()
        return False


class Frame(_Container):
    """A Frame widget that can contain other widgets."""

    def __init__(
        self,
        layout: LayoutType | None = None,
        key: Hashable | None = None,
        width: int | None = None,
        height: int | None = None,
        style: str | None = None,
        borderwidth: int | None = None,
        padding: PaddingType | None = None,
        relief: str = None,
        disabled: bool | None = False,
        rowspan: int | None = None,
        columnspan: int | None = None,
        sticky: str | None = None,
        tooltip: TooltipType | None = None,
        autoframe: bool = True,
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        **kwargs,
    ):
        super().__init__(
            frametype=GUITK.ELEMENT_FRAME,
            key=key,
            width=width,
            height=height,
            layout=layout,
            style=style,
            borderwidth=borderwidth,
            padding=padding,
            relief=relief,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            sticky=sticky,
            tooltip=tooltip,
            autoframe=autoframe,
            kwargs=kwargs,
            valign=valign,
            halign=halign,
            weightx=weightx,
            weighty=weighty,
        )


class LabelFrame(_Container):
    """A Frame widget with a label that can contain other widgets."""

    def __init__(
        self,
        text: str | None = None,
        layout: LayoutType | None = None,
        key: Hashable | None = None,
        width: int | None = None,
        height: int | None = None,
        style: str | None = None,
        borderwidth: int | None = None,
        padding: PaddingType | None = None,
        relief: str = None,
        disabled: bool | None = False,
        rowspan: int | None = None,
        columnspan: int | None = None,
        labelanchor: str | None = None,
        sticky: str | None = None,
        tooltip: TooltipType | None = None,
        autoframe: bool = True,
        weightx: int | None = None,
        weighty: int | None = None,
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        **kwargs,
    ):
        super().__init__(
            frametype=GUITK.ELEMENT_LABEL_FRAME,
            key=key,
            text=text,
            width=width,
            height=height,
            layout=layout,
            style=style,
            borderwidth=borderwidth,
            padding=padding,
            relief=relief,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            labelanchor=labelanchor,
            sticky=sticky,
            tooltip=tooltip,
            autoframe=autoframe,
            kwargs=kwargs,
            valign=valign,
            halign=halign,
            weightx=weightx,
            weighty=weighty,
        )



""" Debug Window that can be used to display debug information """

from __future__ import annotations

from .containers import HStack
from .layout import VLayout
from .tk_text import Output
from .ttk_button import Button
from .ttk_entry import LabelEntry
from .window import Window


class DebugWindow(Window):
    """Debug window that captures stdout/stderr"""

    def __init__(self, output_width=80, output_height=20, **kwargs):
        self._output_width = output_width
        self._output_height = output_height
        super().__init__(**kwargs)

    def config(self):
        self.title = "Debug"
        self.padx = self.pady = 2
        with VLayout():
            with HStack():
                LabelEntry("Filter", key="FILTER_TEXT", width=40),
                Button("Filter", key="FILTER"),
            Output(
                width=self._output_width,
                height=self._output_height,
                key="OUTPUT",
                events=True,
            )

    def handle_event(self, event):
        if event.key in ["FILTER", "OUTPUT"]:
            if filter := self["FILTER_TEXT"].value:
                lines = self["OUTPUT"].value.split("\n")
                lines = [line for line in lines if filter in line]
                self["OUTPUT"].value = "\n".join(lines) + "\n"



""" ttk Scale widget"""

from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable, Union

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Scale"]


_valid_standard_attributes = {
    "class",
    "command",
    "cursor",
    "length",
    "orient",
    "state",
    "style",
    "takefocus",
    "value",
    "variable",
}

_valid_ttk_scale_attributes = {
    "from",
    "to",
} | _valid_standard_attributes


def _interval(from_, to, interval, value, tolerance=1e-9):
    """Clamp value to an interval between from_ and to range"""

    if interval > (to - from_):
        raise ValueError("Invalid increment")

    if value < from_ or value > to:
        raise ValueError("Invalid value")

    if abs(value - from_) < tolerance or abs(value - to) < tolerance:
        return value

    quotient, remainder = divmod(value, interval)
    if remainder < tolerance:
        return quotient * interval

    half_increment = interval / 2
    if remainder > half_increment:
        return interval * (quotient + 1)
    else:
        return interval * quotient


class Scale(BaseWidget):
    """ttk.Scale / slider"""

    def __init__(
        self,
        from_value: float,
        to_value: float,
        value: float | None = None,
        orient: Union[tk.VERTICAL, tk.HORIZONTAL] = tk.VERTICAL,
        interval: float | None = None,
        precision: int | None = None,
        key: Hashable = None,
        target_key: Hashable = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """Initialize a ttk.Scale widget

        Args:
            from_value (float): Minimum value for the scale.
            to_value (float): Maximum value for the scale.
            value (float, optional): Initial value for the scale. Defaults to None (from_value).
            orient (tk.VERTICAL | tk.HORIZONTAL, optional): Orientation of the scale. Defaults to tk.VERTICAL.
            interval (float, optional): Interval between values. Defaults to None (no interval).
            precision (int, optional): Number of decimal places to display. Defaults to None (no rounding).
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            target_key: (Hashable, optional): Unique key for the target widget. Defaults to None.
            default (str | None, optional): Default text for the entry box. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            hscrollbar (bool, optional): Show horizontal scrollbar. Defaults to False.
            weightx (int | None, optional): Weight for horizontal resizing. Defaults to None.
            weighty (int | None, optional): Weight for vertical resizing. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Entry.

        Note:
            Emits EventType.ScaleUpdate event when the scale value changes.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            command=command,
            value_type=tk.DoubleVar,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.widget_type = "ttk.Scale"
        self.key = key or "Scale"
        self.target_key = target_key
        self.kwargs = kwargs

        self.from_ = from_value
        self.to = to_value
        self.orient = orient
        self.interval = interval
        self.precision = precision
        self.initial_value = value

    @property
    def value(self):
        value = self.widget.get()
        if self.interval:
            value = _interval(self.from_, self.to, self.interval, value)
        if self.precision is not None:
            value = round(float(value), self.precision)
        return value

    @value.setter
    def value(self, value):
        self.widget.set(value)

    def _create_widget(self, parent, window: "Window", row, col):
        self.window = window
        event = Event(self, window, self.key, EventType.ScaleUpdate)
        # build arg list for ttk.Scale
        kwargs = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_scale_attributes
        }
        kwargs["variable"] = self._value
        if self.initial_value is not None:
            self._value.set(self.initial_value)

        self.widget = ttk.Scale(
            parent,
            from_=self.from_,
            to=self.to,
            orient=self.orient,
            command=window._make_callback(event),
            **kwargs,
        )
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.ScaleUpdate,
                    command=self._command,
                )
            )

        if self.target_key is not None:
            self.events = True

            def update_target():
                self.window[self.target_key].value = self.value

            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.ScaleUpdate,
                    command=update_target,
                )
            )

        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def scale(self):
        """Return the ttk.Scale widget"""
        return self.widget



""" ttk Progressbar widget """

from __future__ import annotations

import tkinter as tk
import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable, TypeVar

from .basewidget import BaseWidget
from .types import PadType, TooltipType

if TYPE_CHECKING:
    from .window import Window


__all__ = [
    "Progressbar",
    "ProgressBar",
    "PROGRESS_DETERMINATE",
    "PROGRESS_INDETERMINATE",
]

_valid_standard_attributes = {"takefocus", "cursor", "style", "class"}

_valid_ttk_progressbar_attributes = {
    "orient",
    "length",
    "mode",
    "maximum",
    "variable",
    "value",
    "phase",
} | _valid_standard_attributes


Window = TypeVar("Window")

PROGRESS_DETERMINATE = "determinate"
PROGRESS_INDETERMINATE = "indeterminate"


class Progressbar(BaseWidget):
    """ttk.Progressbar"""

    def __init__(
        self,
        key: Hashable | None = None,
        value: float | None = None,
        orient: str = tk.HORIZONTAL,
        length: int = 200,
        mode: str = "determinate",
        maximum: float = 100.0,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        events: bool = True,
        weightx: int | None = None,
        weighty: int | None = None,
        **kwargs,
    ):
        """Initialize a ttk.Progressbar widget

        Args:
            key (Hashable, optional): Key for the widget. Defaults to None.
            value (float, optional): Initial value. Defaults to None.
            orient (str, optional): Orientation of the progress bar. Defaults to tk.HORIZONTAL.
            length (int, optional): Length of the progress bar. Defaults to 200.
            mode (str, optional): Mode of the progress bar. Defaults to "determinate". Can be "determinate" or "indeterminate".
            maximum (float, optional): Maximum value. Defaults to 100.0.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int, optional): Number of columns to span. Defaults to None.
            rowspan (int, optional): Number of rows to span. Defaults to None.
            padx (int, optional): Padding in x direction. Defaults to None.
            pady (int, optional): Padding in y direction. Defaults to None.
            sticky (str, optional): Sticky direction. Defaults to None.
            tooltip (TooltipType, optional): Tooltip for the widget. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to True.
            weightx (int, optional): Weight in x direction. Defaults to None.
            weighty (int, optional): Weight in y direction. Defaults to None.
            **kwargs: Additional keyword arguments to pass to ttk.Progressbar.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            value_type=tk.DoubleVar,
        )
        self.widget_type = "ttk.Progressbar"
        self.key = key or "Progressbar"

        self.orient = orient
        self.length = length
        self.mode = mode
        self.maximum = maximum
        self._initial_value = value
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.tooltip = tooltip
        self.kwargs = kwargs

    @property
    def value(self):
        return self.widget["value"]

    @value.setter
    def value(self, value):
        self._value.set(float(value))

    def _create_widget(self, parent, window: "Window", row, col):
        # Arg list for ttk.ProgressBar
        kwargs_progressbar = {
            k: v
            for k, v in self.kwargs.items()
            if k in _valid_ttk_progressbar_attributes
        }
        kwargs_progressbar["variable"] = self._value
        kwargs_progressbar["orient"] = self.orient
        kwargs_progressbar["length"] = self.length
        kwargs_progressbar["mode"] = self.mode
        kwargs_progressbar["maximum"] = self.maximum
        if self._initial_value is not None:
            self._value.set(self._initial_value)

        self.widget = ttk.Progressbar(parent, **kwargs_progressbar)
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def progressbar(self):
        """Return the ttk.Progressbar widget"""
        return self.widget

    def stop(self):
        self.widget.stop()


class ProgressBar(Progressbar):
    """ttk.Progressbar"""

    pass



"""ttk.Label widget"""

from __future__ import annotations

import sys
import tkinter.ttk as ttk
from tkinter import font
from typing import TYPE_CHECKING, Hashable

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .types import CommandType, PadType, TooltipType, Window
from .utils import load_image

if TYPE_CHECKING:
    from .window import Window

__all__ = ["Label", "LinkLabel", "Linklabel"]

_valid_standard_attributes = {
    "class",
    "compound",
    "cursor",
    "image",
    "style",
    "takefocus",
    "text",
    "textvariable",
    "underline",
    "width",
}

_valid_ttk_label_attributes = {
    "anchor",
    "background",
    "font",
    "foreground",
    "justify",
    "padding",
    "relief",
    "text",
    "wraplength",
} | _valid_standard_attributes


class Label(BaseWidget):
    """ttk.Label widget"""

    def __init__(
        self,
        text: str,
        image: str | None = None,
        key: Hashable | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        weightx: int | None = None,
        weighty: int | None = None,
        **kwargs,
    ):
        """
        Initialize a Label widget.

        Args:
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            text (str): Text to display in the label.
            image: (str, optional): Path to image to display in the label. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            weightx (int | None, optional): Weight of this widget in the horizontal direction. Defaults to None.
            weighty (int | None, optional): Weight of this widget in the vertical direction. Defaults to None.
            **kwargs: Additional keyword arguments are passed to ttk.Entry.
        """
        super().__init__(
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            weightx=weightx,
            weighty=weighty,
        )
        self.widget_type = "ttk.Label"
        self.text = text
        self.key = key or text
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.image = image
        self.kwargs = kwargs

    def _create_widget(self, parent, window: Window, row, col):
        """Create the ttk.Label widget"""

        # Arg list for ttk.Label
        kwargs_label = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_label_attributes
        }

        if self.image:
            self._photoimage = load_image(self.image)
            kwargs_label["image"] = self._photoimage

        self.widget = ttk.Label(
            parent,
            text=self.text,
            **kwargs_label,
        )
        self.widget["textvariable"] = self._value
        self._value.set(self.text)
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )
        if self._disabled:
            self.widget.state(["disabled"])
        return self.widget

    @property
    def label(self):
        """Return the Tk label widget"""
        return self.widget


class LinkLabel(Label):
    """Link label that responds to click"""

    def __init__(
        self,
        text: str,
        key: Hashable | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        **kwargs,
    ):
        """Initialize a LinkLabel widget.

        Args:
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            text (str): Text to display in the label.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to True.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command to execute when clicked. Defaults to None.
            weightx (int | None, optional): Weight of this widget in the horizontal direction. Defaults to None.
            weighty (int | None, optional): Weight of this widget in the vertical direction. Defaults to None.
            **kwargs: Additional keyword arguments are passed to ttk.Entry.
        """
        self.cursor = (
            kwargs.get("cursor") or "pointinghand"
            if sys.platform == "darwin"
            else "hand2"
        )
        super().__init__(
            text=text,
            key=key,
            disabled=disabled,
            rowspan=rowspan,
            columnspan=columnspan,
            padx=padx,
            pady=pady,
            events=events,
            sticky=sticky,
            tooltip=tooltip,
            cursor=self.cursor,
            weightx=weightx,
            weighty=weighty,
        )
        self.widget_type = "guitk.LinkLabel"
        self.text = text
        self.key = key or text
        self.columnspan = columnspan
        self.rowspan = rowspan
        self._command = command
        self.kwargs = kwargs

    def _create_widget(self, parent, window: "Window", row, col):
        self.widget = super()._create_widget(parent, window, row, col)

        event = Event(self, window, self.key, EventType.LinkLabel)
        self.widget.bind("<Button-1>", window._make_callback(event))

        self.widget.configure(cursor=self.cursor)

        if self._command:
            self.events = True
            window._bind_command(
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.LinkLabel,
                    command=self._command,
                )
            )
        return self.widget


class Linklabel(LinkLabel):
    """Non-camel case version"""

    pass



""" ttk Notebook widget """

from __future__ import annotations

import tkinter.ttk as ttk
from typing import TYPE_CHECKING, Hashable, TypeVar

from guitk.constants import GUITK

from .basewidget import BaseWidget
from .events import Event, EventCommand, EventType
from .frame import Frame, _Container, _LayoutMixin, _VerticalContainer
from .layout import get_parent, push_parent
from .types import CommandType, HAlign, PadType, TooltipType, VAlign

if TYPE_CHECKING:
    from .window import Window

# TODO: Add remove() to Notebook and Tab

__all__ = ["Notebook", "NoteBook", "HTab", "VTab"]

_valid_standard_attributes = {
    "width",
    "height",
    "padding",
    "takefocus",
    "cursor",
    "style",
    "class",
}

_valid_ttk_notebook_attributes = _valid_standard_attributes


Window = TypeVar("Window")


class Notebook(_Container):
    """ttk.Notebook widget"""

    def __init__(
        self,
        key: Hashable | None = None,
        tabs: list[HTab] | None = None,
        disabled: bool = False,
        columnspan: int | None = None,
        rowspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = False,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
        **kwargs,
    ):
        """Initialize a Notebook widget.

        Args:
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            tabs: (list[Tab], optional): Tabs to add to the notebook. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command to execute when clicked. Defaults to None.
            weightx (int | None, optional): Horizontal weight. Defaults to None.
            weighty (int | None, optional): Vertical weight. Defaults to None.
            focus (bool, optional): If True, widget will have focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
            **kwargs: Additional keyword arguments are passed to ttk.Entry.


        Note:
            Emits EventType.NotebookTabChanged event.
        """
        super().__init__(
            frametype=GUITK.ELEMENT_FRAME,
            key=None,
            width=None,
            height=None,
            layout=None,
            style=None,
            borderwidth=None,
            padding=0,
            relief=None,
            disabled=disabled,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=False,
            padx=0,
            pady=0,
            weightx=weightx,
            weighty=weighty,
            focus=focus,
        )
        self.widget_type = "ttk.Notebook"
        self.key = key or "Notebook"
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.tabs = tabs or []
        self._command = command
        self.kwargs = kwargs
        self._tab_count = 0

    def _create_widget(self, parent, window: "Window", row, col):
        # Arg list for ttk.Label
        kwargs_notebook = {
            k: v for k, v in self.kwargs.items() if k in _valid_ttk_notebook_attributes
        }

        self.widget = ttk.Notebook(parent, **kwargs_notebook)
        self._grid(
            row=row, column=col, rowspan=self.rowspan, columnspan=self.columnspan
        )

        event_tab_change = Event(self, window, self.key, EventType.NotebookTabChanged)
        self.widget.bind(
            "<<NotebookTabChanged>>", window._make_callback(event_tab_change)
        )

        if self.layout:
            for row in self.layout:
                for tab in row:
                    self.add(tab)

        if self._command:
            self.events = True
            window._bind_command(
                # the actual widget will be a tk widget in form widget=.!toplevel.!frame.!notebook, so it won't match self.widget
                # so set widget=None or _handle_commands won't correctly handle the command
                EventCommand(
                    widget=self,
                    key=self.key,
                    event_type=EventType.NotebookTabChanged,
                    command=self._command,
                )
            )

        if self.width or self.height:
            self.widget.grid_propagate(0)

        if self._disabled:
            self.widget.state(["disabled"])

        return self.widget

    @property
    def current_tab(self):
        """Return the name of the currently selected tab"""
        return self.notebook.tab(self.notebook.select(), "text")

    def add(self, tab: HTab):
        """Add a Tab to the Notebook as new tab"""
        tab_ = tab._create_widget(self.widget, self.window, 0, 0)
        tab.kwargs["text"] = tab.name
        self.notebook.add(tab_, **tab.kwargs)

    def insert(self, pos, tab: HTab):
        """Insert a layout to the Notebook as new tab at position pos"""
        tab_ = tab._create_widget(self.widget, self.window, 0, 0)
        tab.kwargs["text"] = tab.name
        self.notebook.insert(pos, tab_, **tab.kwargs)

    @property
    def notebook(self):
        """Return the ttk.Notebook widget"""
        return self.widget


class NoteBook(Notebook):
    """ttk.Notebook widget"""

    pass


class HTab(_Container):
    """Tab for Notebook widget that arranges its widgets horizontally"""

    def __init__(
        self,
        name=None,
        sticky: str | None = "nsew",
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        **kwargs,
    ):
        """Initialize a horizontal Tab

        Args:
            name (str, optional): Name of the tab. Defaults to None.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            valign (VAlign | None, optional): Vertical alignment of widgets in the tab. Defaults to None.
            halign (HAlign | None, optional): Horizontal alignment of widgets in the tab. Defaults to None.
            **kwargs: Additional keyword arguments are passed to ttk.Frame.
        """

        super().__init__(
            frametype=GUITK.ELEMENT_FRAME,
            key=None,
            width=None,
            height=None,
            layout=None,
            style=None,
            borderwidth=None,
            padding=0,
            relief=None,
            disabled=False,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=True,
            valign=valign,
            halign=halign,
            padx=0,
            pady=0,
        )
        self.name = name
        self.kwargs = kwargs


class VTab(HTab, _VerticalContainer):
    """Tab for Notebook widget that arranges its widgets vertically"""

    def __init__(
        self,
        name=None,
        sticky: str | None = "nsew",
        valign: VAlign | None = None,
        halign: HAlign | None = None,
        **kwargs,
    ):
        """Initialize a vertical Tab

        Args:
            name (str, optional): Name of the tab. Defaults to None.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            valign (VAlign | None, optional): Vertical alignment of widgets in the tab. Defaults to None.
            halign (HAlign | None, optional): Horizontal alignment of widgets in the tab. Defaults to None.
            **kwargs: Additional keyword arguments are passed to ttk.Frame.
        """

        super().__init__(
            frametype=GUITK.ELEMENT_FRAME,
            key=None,
            width=None,
            height=None,
            layout=None,
            style=None,
            borderwidth=None,
            padding=0,
            relief=None,
            disabled=False,
            rowspan=None,
            columnspan=None,
            sticky=sticky,
            tooltip=None,
            autoframe=True,
            padx=0,
            pady=0,
            valign=valign,
            halign=halign,
        )
        self.name = name
        self.kwargs = kwargs



""" HLayout class for use with guitk widgets"""


from __future__ import annotations

import contextlib
import threading
from inspect import currentframe, getmro
from typing import TYPE_CHECKING, Any, Hashable

from ._debug import debug, debug_watch
from .types import HAlign, LayoutType, VAlign

if TYPE_CHECKING:
    from .basewidget import BaseWidget

_current_parent = {}

__all__ = ["HLayout", "VLayout"]


class DummyParent:
    """Dummy class that allows get_parent() to be used whether or not HLayout context manager in use"""

    def _add_widget(self, widget: Any):
        pass


def push_parent(parent: Any):
    """push parent onto stack"""
    global _current_parent
    current_thread = threading.current_thread().ident
    if current_thread not in _current_parent:
        _current_parent[current_thread] = [parent]
    else:
        _current_parent[current_thread].append(parent)


def pop_parent():
    """pop parent off stack"""
    global _current_parent
    current_thread = threading.current_thread().ident
    if current_thread in _current_parent:
        _current_parent[current_thread].pop()


def get_parent() -> Any:
    """Returns current parent"""
    global _current_parent
    current_thread = threading.current_thread().ident
    if current_thread in _current_parent:
        return _current_parent[current_thread][-1]
    return DummyParent()


class HLayout:
    """A Layout manager that aligns widgets horizontally"""

    def __init__(
        self,
        layout: LayoutType = None,
        *,
        valign: VAlign | None = None,
        halign: HAlign | None = None,
    ):
        self._layout_list = layout or []
        self.index = 0
        self.valign = valign
        self.halign = halign
        self.window = None

        # get the caller's instance so we can set the layout
        caller_frame = currentframe().f_back
        with contextlib.suppress(IndexError, KeyError):
            first_arg = caller_frame.f_code.co_varnames[0]
            caller_instance = caller_frame.f_locals[first_arg]
            # determine if the caller is a Window
            # need to use repr() because we can't import Window here without causing a circular import
            if "guitk.window.Window" in repr(getmro(caller_instance.__class__)):
                # HLayout is being used in a Window, so set the Window's layout automatically
                self.window = caller_instance
                self.window.layout = self

    def append(self, widget: BaseWidget):
        """Add a widget to the end of the HLayout"""
        if not self.window:
            # HLayout is not being used in a Window, can't add widget
            raise RuntimeError(
                "Layout must have been created in a Window to add widgets"
            )
        self.window.col_count += 1
        self.window.add_widget(widget, 0, self.window.col_count)

    def remove(self, key_or_widget: Hashable | BaseWidget):
        """Remove widget from layout" and destroy it.

        Args:
            key_or_widget (Hashable | Widget): The key or widget to remove. If a key is given,
                the first widget with that key will be removed.

        Raises:
            RuntimeError: If called and the layout was not created in a Window.
            ValueError: If the widget is not found in the layout.
        """
        if not self.window:
            raise RuntimeError(
                "Layout must have been created in a Window to remove widgets"
            )
        for idx, widget in enumerate(self._layout_list):
            if widget == key_or_widget or widget.key == key_or_widget:
                widget = self._layout_list.pop(idx)
                self.window._forget_widget(widget)
                widget.widget.grid_forget()
                widget.widget.destroy()
                self.window.window.update_idletasks()
                return
        raise ValueError(f"Widget {key_or_widget} not found in Layout")

    def _add_widget(self, widget):
        """Add a widget to the end of the Layout"""
        self._layout_list.append(widget)

    @property
    def layout(self) -> LayoutType:
        """Return the layout list"""
        # if layout manually created, it will be a list of lists
        # otherwise it's a row of widgets, so wrap it in a list
        return (
            self._layout_list
            if self._layout_list and isinstance(self._layout_list[0], (list, tuple))
            else [self._layout_list]
        )

    def __enter__(self):
        push_parent(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pop_parent()
        return False

    def __iter__(self):
        self.index = 0
        return iter(self.layout)

    def __next__(self):
        if self.index >= len(self.layout):
            raise StopIteration
        value = self.layout[self.index]
        self.index += 1
        return value


class VLayout(HLayout):
    """A Layout manager that aligns widgets vertically"""

    @property
    def layout(self):
        return [[w] for w in self._layout_list]

    def append(self, widget: BaseWidget):
        """Add a widget to the bottom of the VLayout"""
        if not self.window:
            # HLayout is not being used in a Window, can't add widget
            raise RuntimeError(
                "VLayout must have been created in a Window to add widgets"
            )
        self.window.row_count += 1
        self.window.add_widget(widget, self.window.row_count, 0)



""" Widget base class """

from __future__ import annotations

import tkinter as tk
from typing import TYPE_CHECKING, Any, Callable, Hashable

from guitk.tkroot import _TKRoot

from ._debug import debug, debug_watch
from .events import Event, EventCommand
from .layout import DummyParent, get_parent
from .types import CommandType, HAlign, PadType, TooltipType, VAlign, ValueType

if TYPE_CHECKING:
    from .frame import _Container
    from .window import Window


class BaseWidget:
    """Basic abstract base class for all tk widgets"""

    def __init__(
        self,
        key: Hashable | None = None,
        disabled: bool = False,
        rowspan: int | None = None,
        columnspan: int | None = None,
        padx: PadType | None = None,
        pady: PadType | None = None,
        events: bool = True,
        sticky: str | None = None,
        tooltip: TooltipType = None,
        command: CommandType | None = None,
        value_type: ValueType | None = None,
        weightx: int | None = None,
        weighty: int | None = None,
        focus: bool = False,
    ):
        """Initialize a widget.

        Args:
            key (Hashable, optional): Unique key for this widget. Defaults to None.
            default (str | None, optional): Default text for the entry box. Defaults to None.
            disabled (bool, optional): If True, widget is disabled. Defaults to False.
            columnspan (int | None, optional): Number of columns to span. Defaults to None.
            rowspan (int | None, optional): Number of rows to span. Defaults to None.
            padx (PadType | None, optional): X padding. Defaults to None.
            pady (PadType | None, optional): Y padding. Defaults to None.
            events (bool, optional): Enable events for this widget. Defaults to False.
            sticky (str | None, optional): Sticky direction for widget layout. Defaults to None.
            tooltip (TooltipType | None, optional): Tooltip text or callback to generate tooltip text. Defaults to None.
            command (CommandType | None, optional): Command callback. Defaults to None.
            value_type (ValueType | None, optional): Type of value to store (for example, tk.StringVar). Defaults to None.
                If None, a tk.StringVar is used.
            weightx (int | None, optional): Weight for horizontal resizing. Defaults to None.
            weighty (int | None, optional): Weight for vertical resizing. Defaults to None.
            focus (bool, optional): If True, widget has focus. Defaults to False.
                Only one widget in a window can have focus.HLayout
        """
        super().__init__()

        self.key = key
        self._disabled = disabled
        self.columnspan = columnspan
        self.rowspan = rowspan
        self.padx = padx
        self.pady = pady
        self.events = events
        self.sticky = sticky or ""
        self.tooltip = tooltip
        self.weightx = weightx
        self.weighty = weighty
        self._focus = focus

        self._command = command
        self._commands = {}

        self.widget_type = None
        self._tk = _TKRoot()
        self.widget: tk.BaseWidget | None = None
        self._value = value_type() if value_type is not None else tk.StringVar()

        # set by _create_widget in inherited classes or the __init_subclass__ method
        self._parent = None
        self.window = None
        self.parent = None

        # used by style()
        self._style_kwargs = {}

        # will store where widget is placed in layout grid (row, column)
        # set by the layout manager
        self._row = None
        self._col = None

        # set to True when _layout creates the widget
        self._has_been_created = False

    @property
    def value(self):
        return self._value.get()

    @value.setter
    def value(self, value):
        self._value.set(value)

    def focus(self):
        self.widget.focus()

    def _grid(self, row, column, rowspan, columnspan):
        sticky = self.sticky or tk.W

        debug("_grid", self, self.widget, row, column, rowspan, columnspan, sticky)
        self.widget.grid(
            row=row,
            column=column,
            columnspan=columnspan,
            rowspan=rowspan,
            sticky=sticky,
            padx=self.padx,
            pady=self.pady,
        )

        if self.padx is not None or self.pady is not None:
            debug(f"{self=} {self.padx=} {self.pady=}")
            self.widget.grid_configure(padx=self.padx, pady=self.pady)

    def bind_event(self, event_name: str, command: Callable[[], Any] | None = None):
        """Bind a tkinter event to widget; will result in an Event being sent to handle_event when triggered.
        Optionally bind command to the event"""
        event = Event(self, self.window, self.key, event_name)
        self.widget.bind(event_name, self.window._make_callback(event))

        if command:
            self.window._bind_command(
                EventCommand(
                    widget=self, key=self.key, event_type=event_name, command=command
                )
            )

    @property
    def state(self) -> bool:
        return self.widget["state"]

    @property
    def disabled(self) -> bool:
        return self.widget["state"] == "disabled"

    @disabled.setter
    def disabled(self, value: bool) -> None:
        self.widget["state"] = "disabled" if value else "normal"

    def style(self, **kwargs) -> BaseWidget:
        """Configure the widget style

        Args:
            **kwargs -- style options

        Returns: Widget instance (self)

        Note:
        The specific style options available depend on the widget type.
        For example, to configure the foreground color to blue of a Label widget:
            Label("Hello").style(foreground="blue")

        Refer to the tk documentation for the specific widget type for more information.
        https://tkdocs.com/index.html

        style() returns the widget instance, so it can be chained with other methods,
        for example:
            Label("Hello").style(foreground="blue").font(size=20)
        """
        self._style_kwargs |= kwargs
        self._configure()
        return self

    def font(
        self,
        font: str | None = None,
        family: str | None = None,
        size: int | None = None,
        weight: str | None = None,
        slant: str | None = None,
        underline: bool | None = None,
        overstrike: bool | None = None,
    ) -> BaseWidget:
        """Configure the widget font

        Args:
            font -- font specifier (name, system font, or (family, size, style)-tuple)
            family -- font 'family', e.g. Courier, Times, Helvetica
            size -- font size in points
            weight -- font thickness: NORMAL, BOLD
            slant -- font slant: ROMAN, ITALIC
            underline -- font underlining: false (0), true (1)
            overstrike -- font strikeout: false (0), true (1)

        Returns: Widget instance (self)

        Note:
            The font parameter is a string of the form "family size style", where style is a string
            containing any combination of "bold", "italic", "underline", and "overstrike" separated by spaces.
            The size may be specified as an integer, or as a floating point number followed by the unit "p" or "P".
            The unit "p" or "P" stands for points, 1/72 of an inch, and is the standard unit used by X fonts.
            If font is specified, then the other font-related options (size, weight, slant, underline, overstrike)
            are ignored.
        """
        locals_ = locals()
        locals_.pop("self")
        if font_kwargs := {k: v for k, v in locals_.items() if v is not None}:
            self._style_kwargs["font"] = tk.font.Font(**font_kwargs)
        self._configure()
        return self

    # def valign(self, valign: VAlign | None = None) -> Widget:
    #     """Set valign the widget

    #     Args:
    #         valign (VAlign | None): Vertical alignment

    #     Returns: Widget instance (self)
    #     """

    #     return self

    @debug_watch
    def destroy(self) -> None:
        """Remove the widget from the layout and destroy it.

        Note:
            You should call this destroy() method and not the tkinter destroy() method
            so that the widget is removed from its parent and necessary bookkeeping is done
            in the Window class.
        """
        debug("destroy", self, self.parent)
        self.parent.remove(self)

    def replace(self, widget: BaseWidget) -> BaseWidget:
        """Replace widget with another widget.
        This destroys the parent widget and replaces it with the new widget in the layout.

        Args;
            widget (Widget): New widget

        Returns: Widget instance
        """
        self.widget.grid_forget()
        self.destroy()
        self.parent._insert_widget_row_col(widget, row=self._row, col=self._col)
        return widget

    def _set_parent_window(self, tk_parent: tk.BaseWidget, window: Window):
        """Called during widget creation to set the widget attributes"""
        self.window = window
        self._parent = tk_parent

    def _set_row_col(self, row: int, col: int):
        """Called during widget creation or configuration to set the widget attributes"""
        self._row = row
        self._col = col

    def _configure(self):
        """Configure the widget after it has been created.
        This allows font(), style() to be used before or after creation.
        """
        if self._has_been_created:
            self.parent._configure_widget(
                self, self._parent, self.window, self._row, self._col
            )

    def __init_subclass__(subclass, *args, **kwargs):
        """Ensure that all widgets are added to the parent layout"""

        # This rewrites the __init__ method of the subclass to add the widget to the parent layout
        super().__init_subclass__(*args, **kwargs)

        @debug_watch
        def new_init(self, *args, init=subclass.__init__, **kwargs):
            init(self, *args, **kwargs)
            if subclass is type(self):
                # only do this for the bottom grandchild class
                # in the case of subclassed widgets
                self.parent = get_parent()
                if isinstance(self.parent, DummyParent):
                    raise RuntimeError(
                        "Widget must created within a Layout or Container"
                    )
                debug(f"{self.parent=} {self=}")
                self.parent._add_widget(self)

        subclass.__init__ = new_init
